/*
 * Created on Mar 4, 2006
 *
 */
package javaMIDISoundServer;

import java.net.*;

import javax.sound.midi.MidiChannel;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Synthesizer;

/**
 * @author dstulken
 *
 * JaMISS.java - Java MIDI Sound Server for ISU HCI research
 * (c) 2006 David Stulken, Iowa State University
 *
 * Functions:
 * 		1) Starts a Java MIDI synthesizer
 * 		2) Accepts connections on a given port
 * 		3) Allows remote control of sound creation & its parameters
 *
 * In addition, there is a diagnostic mode.  In this mode, the program accepts connections, and simply prints the data it receives.
 *
 * Usage:
 *  #java -jar jaMISS [arguments]
 *
 *
 * Command line arguments:
 *  -p [port]	Specifies the port to listen on (Default is 8006)
 *  -d			Enable diagnostic mode (No synth used, only prints data received)
 *  --info		Returns program info (what it does, packet format, etc...)
 *  --help		Returns program usage & parameters
 *
 * JaMISS is very robust.
 * - If the port argument is omitted, the default port (8006) is used.
 * - Parameters can be entered in any order
 * - Boundary conditions (ports, etc) are checked.
 * - Parameters are evaluated in logical order: '--help' in any location will override any other parameters
 *
 *
 */
public class JaMISS extends Thread {
	/*
	 	JaMISS Server  (yes, that is redundant)

	 	Waits for a request
	 	Spawns a thread to handle the request
	 	Main thread goes back to waiting for requests
	 */

	//MIDI variables:
	boolean midi_Active=false;	//initially, no note is playing
	int midi_Channel=-1;		//initially, no channel
	int midi_Note=-1;			//initially, no note
	int midi_Instrument=-1;		//initially, no instrument
	int midi_Velocity=-1;		//initially, no velocity.
	static Synthesizer softSynth=null;				//software Synthesizer instance
	static MidiChannel[] softSynthChannels = null; //array of synth channels (capabilities)


	public static void main(String[] args) {

	//declare major variables (others will be locally declared inline)
	//input parameters
	boolean param_diag = false;
	boolean param_help = false;
	boolean param_info = false;
	boolean param_none = true;
	int port = 8006;  //set to default of 8006.  Will be set to -1 if there is an error in the command line arguments


	//print startup message
	System.out.println("========================================================================");
	System.out.println("Java MIDI Sound Server started ...                      (ctrl-c to quit)");
	System.out.println("========================================================================");

	String OS=JaMISS.detectOS();
	if (OS=="WIN"){
		System.out.println("  >> WINDOWS WARNING:");
		System.out.println("  >> If no sound is heard, you may need to install a MIDI Soundbank.");
		System.out.println("  >> Downloads and installation instructions can be found at:");
		System.out.println("  >> http://java.sun.com/products/java-media/sound/soundbanks.html");
	}


	if (args.length>0){
		//if no parameters, set "no parameters" variable to true
		param_none=false;
	}

	//if (param_none){
	//	//if no parameters specified, show help. (by assigning help to be true)
	//	param_help=true;
	//}

	//**************************************************************************************************************
	//parse input arguments
	//**************************************************************************************************************
	for (int argX=0; argX<args.length; argX++)
	{
		if (args[argX].compareTo("-d")==0)
		{
			param_diag=true;
		}
		if (args[argX].compareTo("--help")==0)
		{
			param_help=true;
		}
		if (args[argX].compareTo("--info")==0)
		{
			param_info=true;
		}


		if (args[argX].compareTo("-p")==0)
		{
			//if a port argument is specified (or an attempt made)
			if (argX+1>args.length)	//if '-p' is last item entered (ie: they forgot the port)
			{
				port=-1;	//set port to the error value
			}
			else
			{
				//NOTE: THIS SECTION CAN BE DONE MORE EFFICIENTLY - CHANGE WHEN TIME PERMITS!!
				// use try/catch with parseInt, catch numberformat exception


				//if the next entry is valid array item, check if it really is a port
				String tempPort=args[argX+1];	//set tempPort to next array item

				//next, check if it is NOT a valid number.
				for (int portNumX=0; portNumX<tempPort.length(); portNumX++)
				{
					//steps through each character in port number string
					char tP=tempPort.charAt(portNumX);	//assign char tP to current position in port number field

					//checking if any character is invalid (by seeing if any IS a legal value, then inverting)
					if (!(tP=='0' || tP=='1' || tP=='2' || tP=='3' || tP=='4' || tP=='5' || tP=='6' || tP=='7' || tP=='8' || tP=='9'))
					{
						//If we get here, it is NOT a number, and is therefore not a valid command line argument.
						port=-1;	//set port to the error value
					}
					else
					{
						//if we get here, the character must be valid.  Do nothing at this point.
					}
				}
				//if we get here, and port does not equal -1, we must have a valid port number, so assign it.
				if (port!=-1)
				{
					port=Integer.parseInt(args[argX+1]);
					argX++; //advance argX counter to skip next value (which we just read)

					System.out.println("** Debug: FIND RANGE OF ACCEPTABLE PORTS AND BOUND INPUT RANGE!**");
					if (false) //set to check condition!!
					{
						port=-1;  //set port back to -1
					}

				}

			} //end of port number check/assign

		} //end of "-p" checking

	} //end of input parsing




	//**************************************************************************************************************
	//Decide what to do based on input arguments
	//**************************************************************************************************************
	// if 'help' requested, ignore other command line errors
	// if port is -1, we had an error


	//check if help requested
	if (param_help==true)
	{
		printHelp();	//call printHelp method
		System.exit(0); //exit
	}

	//check if info requested
	if (param_info==true)
	{
		printInfo();	//call printHelp method
		System.exit(0); //exit
	}

	//at this point, we can go ahead.  The decision to start the actual MIDI stuff will be evaluated as needed.
	if (param_diag==true)
	{
		System.out.println(">> Running in diagnostic mode.");
	}
	else
	{
		//not in diagnostic mode, so start midi system
		try
		{
			softSynth = MidiSystem.getSynthesizer();
		}
		catch (MidiUnavailableException e)
		{
			//do something different?
			System.out.println(">>> MIDI Error 1 - Can't find synth.");
			System.out.println("System Exiting.");
			//System.err.println(e);
			System.exit(1);
		}
		System.out.println(">> Found Synth:"+softSynth);
		try
		{
			softSynth.open();
		}
		catch (MidiUnavailableException e)
		{
			//do something different?
			System.out.println(">>> MIDI Error 2 - Can't open synth.");
			System.out.println("System Exiting.");
			//System.err.println(e);
			//e.printStackTrace();
			System.exit(1);
		}
		System.out.println(">> Synth opened.");
		//Find capabilities of synth
		//  (query the synth for available channels)
		softSynthChannels=softSynth.getChannels();
		System.out.println(">> Synth capabilities determined.");
		//note: if we reach here, then all was successful.  Can assume this later in program without any consequences.
	}

	//	Check the port value.  If invalid, it will be -1.  If valid, it will be the new value.  If not entered, it will be 8006.
	if (port==-1)
	{
		System.out.println("*** Error! Invalid port number entered! ***");
		System.out.println("Please check your program parameters and try again.");
		System.exit(0);	//exit program
	}
	//didn't fail, so port is valid

	//create server socket
	ServerSocket serverA =null;

	//open socket connection
    boolean noSocket=true;
    while (noSocket)
    {
    	try
		{
    		//initialize socket
    		serverA = new ServerSocket(port);
    		noSocket=false; //exit out of loop
    	}
    	catch(Exception s)
		{
    		//print error message
    		System.out.println("*** Socket exception. Please check your port number. ***");
    		System.err.println(s);
    	}
    }
	//System.out.println(">> Server socket created.");

	//System.out.println(">> Listening for connections...");
	//System.out.println("------------------------------------------------------------------------");

	boolean qHit=false;
	//loop forever (accept an infinite number of connections)
	while(qHit==false)
	{
		//start the server listening on the previously established port
		try
		{
			//start listening
			Socket socketA = serverA.accept();

			//System.out.println(">> Connection received!");

			//send connection to another thread
			new ServerThread(socketA, param_diag).start();

		}
		catch(Exception e)
		{
			//print generic error message
    		System.out.println("*** Socket exception. Please check your port number. ***");
			System.err.println(e);
		}
    }//end of while loop

	/**
	 *commented out, as this is unreachable code.  this should be fixed... add a quit mechanism to program
	 *
	 *
	//done taking connections, so close the server
	try{
		serverA.close();
	}
    catch(Exception f) {
    	//print generic error message
    	System.out.println("*** Server Error! ***");
    	System.err.println(f);
    }

	//wait half a second to let all threads finish
	try{
		Thread.sleep(500);
	}
	catch (InterruptedException e){
		//exception handler. Intentionally does nothing.
	}

	//print "goodbye" message
	System.out.println("\nThanks for using JaMISS!");
	**/

	}//end of main method

	public static void printHelp()
	{
		//prints help screen
		System.out.println("JaMISS Help:");
		System.out.println("-------------------------------------------------------");
		System.out.println("Usage:");
		System.out.println(" #java -jar jaMISS [arguments]");
		System.out.println();
		System.out.println("Command line arguments:");
		System.out.println("  -p [port]     Specifies the port to listen on (Default is 8006)");
		System.out.println("  -d            Enable diagnostic mode (No synth used, only prints data received)");
		System.out.println("  --info        Returns program info (what it does, packet format, etc...)");
		System.out.println("  --help        Returns program usage & parameters");
		System.out.println();
		System.out.println("JaMISS inputs are fairly robust.");
		System.out.println(" - If the port argument is omitted, the default port (8006) is used.");
		System.out.println(" - Parameters can be entered in any order");
		System.out.println(" - Boundary conditions (ports, etc) are checked for validity");
		System.out.println(" - Parameters are evaluated in logical order");
		System.out.println("       (Example: '--help' in any location will override any other parameters)");

	}

	public static void printInfo()
	{
		//prints help screen
		System.out.println("JaMISS Info:");
		System.out.println("-------------------------------------------------------");
		System.out.println("JaMISS.java - Java MIDI Sound Server for ISU HCI research");
		System.out.println();
		System.out.println("Functions:");
		System.out.println("     1) Starts a Java MIDI synthesizer");
		System.out.println("     2) Accepts connections on a given port");
		System.out.println("     3) Allows remote control of sound creation & its parameters");
		System.out.println();
		System.out.println("In addition, there is a diagnostic mode.");
		System.out.println("    In this mode, the program accepts connections, and simply prints the data it receives.");
		System.out.println();
		System.out.println("Communications:");
		System.out.println("    JaMISS accepts socket connections on the specified port.  The data format is as follows:");
		System.out.println("        JaMISS,[command]");
		System.out.println("          or");
		System.out.println("        JaMISS,[channel],[note],[inistrument],[velocity]");
		System.out.println("    'JaMISS' is a header, and should precede every message string.");
		System.out.println("    [command] can be either 'off' which turns off the note playing, or 'quit' which shuts down the server.");
		System.out.println("    [channel] is the range 0-15 (MIDI channels 1-16). 10 should be avoided (drum track only)");
		System.out.println("    [note] is the range 0-127, but note that values too low are inaudible and values too high are annoying.");
		System.out.println("    [instrument] is the range 0-127, and selects the General MIDI instrument.");
		System.out.println("        (examples: 0 is grand piano which decays, 81 is sqare wave which has infinite sustain.)");
		System.out.println("    [velocity] is the intensity of the instrument. It may or may not be the same as volume, depending on the sound.");

	}

	public static String detectOS(){

		String OS=null;
		OS=System.getProperty("os.name").toUpperCase();
		System.out.println(">> Detected OS: "+ OS);

		if ((OS.substring(0,3)).compareTo("WIN") == 0){
			OS="WIN";
		}


		return OS;
	}

	public static void updateSound(String command, int channel, int note, int instrument, int velocity)
	{
		//This is the method that actually changes sound parameters. It is presently called from the server thread.

		//NOTE: This currently does not have any mutex locking- it probably should!!

		if (command.compareTo("off")==0){
			//optional command: off
			//System.out.println("  >> Server received 'off' message.  Killing all active notes.");
			softSynthChannels[channel].allNotesOff();
		}
		else if (command.compareTo("quit")==0){
			//optional command: quit
			//System.out.println("  >> Server received 'quit' message.  Shutting down.");
			softSynthChannels[channel].allNotesOff();
			//Done, so close the synth
			softSynth.close();
			//System.out.println(">> Synth closed.");
			//System.out.println("\nHave a nice day!");
			System.exit(0);
		}
		else{
			//no optional command

			System.out.println("  >> Updating sound parameters:\n\t\tchannel:"+channel+",  note:"+note+",  instrument:"+instrument+",  velocity:"+velocity);

			//NOTE: WE CAN STILL RECEIVE BAD DATA HERE!! DO MORE ERROR CHECKING!

			if (channel<0 || channel>15 || note<0 || note>127 || instrument<0 || instrument>127 || velocity<0 || velocity>127){
				//if parameters out of range...
				//System.out.println("  >> Parameters out of range.  Please double check your messages!");
				//System.out.println("  >> Sound not updated.");
			}
			else{
				//parameters in range, so update.
				softSynthChannels[channel].allNotesOff();				//kill all previously playing sounds
				softSynthChannels[channel].programChange(instrument);   //change to the selected instrument
				softSynthChannels[channel].noteOn(note,velocity);		//trigger our new note at the given velocity
				//System.out.println("  >> Sound updated!");
			}

		}
	} //end of updateSound method
}




